#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Wrapper around the TMK+PDQF signal type.
"""

import typing as t

from ..descriptor import SimpleDescriptorRollup, ThreatDescriptor
from . import signal_base


class VideoTmkPdqfSignal(signal_base.SimpleSignalType):
    """
    TMK+PDQF is an open-source, constant length hash for videos.

    Video MD5s are a constant length hash, which makes them simple to
    implement a lookup against an index of hashes, but are sensitive to minor
    changes in the video content (even a single pixel changed changes the MD5).
    TMK+PDQF has a notion of "distance" which can be used to tell when content
    is similar, if not identical.
    """

    INDICATOR_TYPE = "HASH_TMK"
    TYPE_TAG = "media_type_long_hash_video"

    # TODO at least hash distance functionality



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Wrapper around the video content type.
"""
import typing as t

from ..signal_type import raw_text, trend_query, url, md5, video_tmk_pdqf
from ..signal_type.signal_base import SignalType
from .content_base import ContentType


class VideoContent(ContentType):
    """
    Content representing a sequence of images, giving the illusion of motion.

    Examples might be:
    * mp4
    * avi
    * gif animations
    """

    @classmethod
    def get_signal_types(cls) -> t.List[t.Type[SignalType]]:
        return [md5.VideoMD5Signal, video_tmk_pdqf.VideoTmkPdqfSignal]



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
A command to fetch datasets from ThreatExchange based on the collab config
"""

import argparse
import collections
import concurrent.futures
import datetime
import inspect
import pathlib
import time
import typing as t
import urllib.parse

from ..api import ThreatExchangeAPI
from ..collab_config import CollaborationConfig
from ..content_type import meta
from ..dataset import Dataset, FetchCheckpoint
from ..descriptor import ThreatDescriptor
from ..signal_type import signal_base
from . import command_base


class FetchType:
    """Typed helper determine_fetch_type."""

    def __init__(self, from_timestamp: float) -> None:
        self.from_timestamp = from_timestamp

    @property
    def is_full(self) -> bool:
        return not self.from_timestamp

    @property
    def is_incremental(self) -> bool:
        return not self.is_full

    @classmethod
    def Full(cls) -> "FetchType":
        return cls(0.0)

    @classmethod
    def Incremental(cls, from_timestamp: float) -> "FetchType":
        return cls(from_timestamp)


class TagFetchCommand(command_base.Command):
    """
    Download content from ThreatExchange to disk.

    Using the CollaborationConfig, identify ThreatDescriptors that
    correspond to a single collaboration, and store them in the state
    directory.

    You can then use the match command to search against this directory for
    the simple content, or you can use the produced files (which by default
    live in your home directory with a name based on the collaboration, but
    can be overridden with the --state-dir argument) to load into your own
    infrastructure to more efficiently match against the downloaded hashes.

    The exact format of each file is determined by the implementation of the
    signal types, but or usually optimized for easy re-use, such as .csv or
    .tsv.
    """

    # Enforced by the endpoint
    MAX_DESCRIPTOR_FETCH_SIZE = 20
    DEFAULT_REFETCH_SEC = 3600 * 24 * 7  # 1 week
    UP_TO_DATE_SEC = 60  # 1 minute
    PROGRESS_PRINT_INTERVAL_SEC = 30

    @classmethod
    def init_argparse(cls, ap) -> None:
        ap.add_argument(
            "--sample",
            action="store_true",
            help="Only fetch a sample of data instead of the whole dataset",
        )
        ap.add_argument(
            "--clear",
            action="store_true",
            help="Don't fetch anything, just clear the dataset",
        )
        ap.add_argument(
            "--full", action="store_true", help="Force a full refresh of data."
        )
        ap.add_argument(
            "--until-timestamp",
            type=int,
            default=0,
            help=(
                "Instead of fetching all the data, "
                "incrementally fetch up to this time"
            ),
        )
        ap.add_argument(
            "--only-signals",
            "-o",
            nargs="+",
            metavar="signal",
            choices=sorted(meta.get_signal_types_by_name()),
            help="Only download the specified signal types.",
        )
        ap.add_argument(
            "--not-signals",
            "-n",
            nargs="+",
            metavar="signal",
            choices=sorted(meta.get_signal_types_by_name()),
            help="Don't fetch the specified signal types.",
        )

    def __init__(
        self,
        sample: bool,
        clear: bool = False,
        until_timestamp: int = 0,
        full: bool = False,
        only_signals: t.Collection[str] = (),
        not_signals: t.Collection[str] = (),
    ) -> None:
        """Has default arguments because it's called by match command"""
        self.sample = sample
        self.clear = clear

        now = time.time()
        self.until_timestamp = min(float(until_timestamp or now), now)
        self.force_full = full
        self.force_incremental = bool(until_timestamp) and not full
        # Cause first print to be after ~5 seconds
        self.last_update_printed = now - self.PROGRESS_PRINT_INTERVAL_SEC + 5

        if only_signals or not_signals:
            # TODO - this is fixable by storing checkpoints per signal type
            self.stderr(
                "Ignoring some signal types will lead to only part of",
                "the dataset being fetched until fixed with --full.",
            )

        only_signals = only_signals or meta.get_signal_types_by_name()
        not_signals = not_signals or ()
        self.signal_types_by_name = {
            name: signal()
            for name, signal in meta.get_signal_types_by_name().items()
            if name in only_signals and name not in not_signals
        }

    def determine_fetch_type(self, checkpoint: FetchCheckpoint) -> FetchType:
        """Based on the checkpoint and options, determine what fetch we are doing"""
        if self.force_full or self.sample:
            return FetchType.Full()

        fetch_from_timestamp = checkpoint.last_fetch

        if not checkpoint.last_fetch:
            return FetchType.Full()

        if (
            not self.force_incremental
            and time.time() - checkpoint.last_full_fetch > self.DEFAULT_REFETCH_SEC
        ):
            self.stderr("It's been a long time since a full fetch, forcing one now.")
            return FetchType.Full()

        return FetchType.Incremental(checkpoint.last_fetch)

    def execute(self, api: ThreatExchangeAPI, dataset: Dataset) -> None:
        if self.clear:
            dataset.clear_cache()
            return

        fetch_type = self.determine_fetch_type(dataset.get_fetch_checkpoint())

        if fetch_type.is_incremental:
            if fetch_type.from_timestamp >= self.until_timestamp:
                self.stderr("Already up-to-date.")
                return
            start_time_str = datetime.datetime.utcfromtimestamp(
                int(fetch_type.from_timestamp)
            ).strftime("%Y-%m-%d %H:%M:%S")
            self.stderr(
                f"Doing an incremental update from {start_time_str}. ",
                (
                    "This will miss some updates and deletes of previously fetched data, "
                    "use --full to force a refetch."
                ),
                sep="\n",
            )
            # Note that merging into this is error prone - tagging and untagging a descriptor
            # can move its position in the fetch, leading to the following
            # 1. Descriptor1 (D1) => Tag1 (T1)
            # 2. Fetch stores D1: T1
            # 3. D1 untagged, re-tagged with T2
            # 4. fetch load D1: T1
            # 5. fetch finds D1: T2
            # 6. Naive merge yield D1: T1, T2
            #
            # We're relying on full fetching to fix this eventually (as well as deletes)
            dataset.load_cache(self.signal_types_by_name.values())

        with concurrent.futures.ThreadPoolExecutor(max_workers=100) as id_fetch_pool:

            seen_td_ids = set()

            counts = collections.Counter()
            errors = collections.Counter()

            tags_to_fetch = dataset.config.labels
            only_first_fetch = False

            if self.sample:
                if dataset.config.sample_tag:
                    tags_to_fetch = [dataset.config.sample_tag]
                else:
                    only_first_fetch = True

            # TODO - Write a checkpoint file on descriptors, potentially resume from that file
            #        if we exit

            def consume_descriptors(dq: collections.deque) -> int:
                """Process descriptors in order"""
                item = dq.popleft()
                # TODO - consider a timeout
                descriptors = item.result()
                for descriptor in descriptors:
                    any_match = False
                    for signal_name, signal_type in self.signal_types_by_name.items():
                        is_match = False
                        try:
                            is_match = signal_type.process_descriptor(descriptor)
                        except Exception:
                            errors[signal_name] += 1
                        if is_match:
                            any_match = True
                            counts[signal_name] += 1
                    if any_match:
                        counts["all"] += 1
                now = time.time()
                if now - self.last_update_printed >= self.PROGRESS_PRINT_INTERVAL_SEC:
                    self.last_update_printed = now
                    self.stderr(f"Processed {counts['all']}...")
                return len(descriptors)

            for tag_name in tags_to_fetch:
                tag_id = api.get_tag_id(tag_name)
                if not tag_id:
                    continue
                pending_futures = collections.deque()
                remainder_td_ids = collections.deque()
                query = _TagQueryFetchCheckpoint(
                    api, tag_id, fetch_type.from_timestamp, self.until_timestamp
                )

                # Query tags in order on a single thread to prevent overfetching ids
                while query:
                    ids = [i for i in query.next() if i not in seen_td_ids]
                    seen_td_ids.update(ids)
                    remainder_td_ids.extend(ids)
                    while len(remainder_td_ids) >= self.MAX_DESCRIPTOR_FETCH_SIZE:
                        batch = [
                            remainder_td_ids.popleft()
                            for _ in range(self.MAX_DESCRIPTOR_FETCH_SIZE)
                        ]
                        pending_futures.append(
                            id_fetch_pool.submit(
                                self._fetch_descriptors,
                                api,
                                batch,
                                dataset.config.labels,
                            )
                        )
                    if only_first_fetch:
                        break
                    # Consume descriptor data as it becomes available, or if we get too far ahead
                    # to try and avoid a memory explosion
                    while pending_futures and (
                        len(pending_futures) > 200 or pending_futures[0].done()
                    ):
                        consume_descriptors(pending_futures)
                        # TODO Some kind of checkpointing behavior
                # Submit any stragglers
                if remainder_td_ids:
                    pending_futures.append(
                        id_fetch_pool.submit(
                            self._fetch_descriptors,
                            api,
                            list(remainder_td_ids),
                            dataset.config.labels,
                        )
                    )
                while pending_futures:
                    consume_descriptors(pending_futures)

        if fetch_type.is_full and not counts:
            raise command_base.CommandError(
                "No items fetched! Something wrong?", returncode=3
            )
        del counts["all"]  # Not useful for final display

        for signal_name, signal_type in self.signal_types_by_name.items():
            if signal_name not in counts:
                continue
            dataset.store_cache(signal_type)
            print(f"{signal_name}: {counts[signal_name]}")
        if errors:
            self.stderr("\nSome signal types had errors:")
            for signal_name, error_count in sorted(errors.items(), key=lambda t: -t[1]):
                self.stderr(f"  {signal_name}: {error_count}")
        if not self.sample:
            dataset.record_fetch_checkpoint(
                self.until_timestamp or self.start_time, fetch_type.is_full
            )

    def _fetch_descriptors(
        self, api: ThreatExchangeAPI, td_ids: t.List[int], collab_labels: t.List[str]
    ) -> t.List[ThreatDescriptor]:
        """Do the bulk ThreatDescriptor fetch"""
        return [
            ThreatDescriptor.from_te_json(ThreatDescriptor.MY_APP_ID, td_json)
            for td_json in api.get_threat_descriptors(td_ids)
        ]


class _TagQueryFetchCheckpoint:
    def __init__(
        self, api: ThreatExchangeAPI, tag_id: int, since: float = 0, until: float = 0
    ) -> None:
        self.api = api
        query = {"access_token": api.api_token, "limit": 1000, "fields": "id,type"}
        # Surprise! The endpoint accepts since/until but then ignores them!
        # You must know the secret command words are "tagged_since" and "tagged_until"
        if since:
            query["tagged_since"] = int(since)
        if until:
            query["tagged_until"] = int(until)
        query_str = urllib.parse.urlencode(query)
        # TODO - clean this up, consider making cursoring from a helper object in api
        self._next_url = f"{api._TE_BASE_URL}/{tag_id}/tagged_objects/?{query_str}"

    def __bool__(self) -> bool:
        return bool(self._next_url)

    def next(self) -> t.Dict[id, t.Any]:
        response = self.api.get_json_from_url(self._next_url)
        self._next_url = response.get("paging", {}).get("next")
        return [d["id"] for d in response["data"] if d["type"] == "THREAT_DESCRIPTOR"]



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
A wrapper around multi-stage ThreatExchange operations.

Includes simple matching and writing back. Useful for quickly validating new
sources of ThreatExchange data. A possible template for a native
implementation in your own architecture.

This helper heavily relies on a config file to provide consistent behavior
between stages, and a state file to store hashes.
"""

import argparse
import inspect
import os
import os.path
import pathlib
import re
import sys
import typing as t

from .. import descriptor
from ..api import ThreatExchangeAPI
from ..collab_config import CollaborationConfig
from ..dataset import Dataset
from . import (
    command_base as base,
    fetch,
    tag_fetch,
    label,
    match,
    dataset_cmd,
    hash_cmd,
)


def get_subcommands() -> t.List[t.Type[base.Command]]:
    return [
        fetch.FetchCommand,
        match.MatchCommand,
        label.LabelCommand,
        dataset_cmd.DatasetCommand,
        hash_cmd.HashCommand,
        tag_fetch.TagFetchCommand,
    ]


def get_argparse() -> argparse.ArgumentParser:
    ap = argparse.ArgumentParser(
        description=__doc__, formatter_class=argparse.RawDescriptionHelpFormatter
    )
    ap.add_argument(
        "--config",
        "-c",
        type=argparse.FileType("r"),
        help="a ThreatExchange collaboration config",
    )
    ap.add_argument(
        "--app-token",
        "-a",
        metavar="TOKEN",
        help="the App token for ThreatExchange",
    )
    ap.add_argument(
        "--state-dir",
        "-s",
        type=_verify_directory,
        metavar="DIR",
        help="the directory with the config state",
    )
    ap.add_argument(
        "--fb-threatexchange-endpoint",
        "-E",
        # For facebook developers testing new APIs, allow pointing
        # to internal endpoints
        help=argparse.SUPPRESS,
    )
    subparsers = ap.add_subparsers(title="verbs", help="which action to do")
    for command in get_subcommands():
        command.add_command_to_subparser(subparsers)

    return ap


def execute_command(namespace) -> None:
    if not hasattr(namespace, "command_cls"):
        get_argparse().print_help()
        return
    command_cls = namespace.command_cls
    try:
        # Init API
        api = ThreatExchangeAPI(
            get_app_token(namespace.app_token),
            endpoint_override=namespace.fb_threatexchange_endpoint,
        )
        # Init state library (needs to be refactored)
        descriptor.ThreatDescriptor.MY_APP_ID = api.api_token.partition("|")[0]
        # Init collab config
        cfg = init_config_file(namespace.config)
        # "Init" dataset
        dataset = Dataset(cfg, namespace.state_dir)
        command_argspec = inspect.getfullargspec(command_cls.__init__)
        arg_names = set(command_argspec[0])
        # Since we didn't import click, use hard-to-debug magic to init the command
        command = command_cls(
            **{k: v for k, v in namespace.__dict__.items() if k in arg_names}
        )
        command.execute(api, dataset)
    except base.CommandError as ce:
        print(ce, file=sys.stderr)
        sys.exit(ce.returncode)
    except KeyboardInterrupt:
        # No stack for CTRL+C
        sys.exit(130)


def get_app_token(cli_option: str = None) -> str:
    """Get the API key from a variety of fallback sources"""

    file_loc = pathlib.Path("~/.txtoken").expanduser()
    environment_var = "TX_ACCESS_TOKEN"
    token = ""
    source = ""
    if cli_option:
        source = "cli argument"
        token = cli_option
    elif os.environ.get(environment_var):
        source = f"{environment_var} environment variable"
        token = os.environ[environment_var]
    elif file_loc.exists() and file_loc.read_text():
        source = file_loc
        token = file_loc.read_text()
    else:
        raise base.CommandError(
            (
                "Can't find App Token, pass it in using one of: \n"
                "  * a cli argument\n"
                f"  * in the environment as {environment_var}\n"
                f"  * in a file at {file_loc}\n"
                "https://developers.facebook.com/tools/accesstoken/"
            ),
            2,
        )
    token = token.strip()
    if not is_valid_app_token(token):
        raise base.CommandError(
            f"Your current app token (from {source}) is invalid.\n"
            "Double check that it's an 'App Token' from "
            "https://developers.facebook.com/tools/accesstoken/",
            2,
        )
    return token


def is_valid_app_token(token: str) -> bool:
    """Returns true if the string looks like a valid token"""
    return bool(re.match("[0-9]{8,}(?:%7C|\\|)[a-zA-Z0-9_\\-]{20,}", token))


def init_config_file(cli_provided: t.IO = None) -> CollaborationConfig:
    """Initialize the collaboration file from a variety of sources"""
    if cli_provided is not None:
        return CollaborationConfig.load(cli_provided)
    path_order = ("te.cfg", "~/te.cfg")
    for loc in path_order:
        path = pathlib.Path(loc).expanduser()
        if path.exists():
            break
    else:
        print(
            (
                "Looks like you haven't set up a collaboration config, "
                "so using the sample one against public data"
            ),
            file=sys.stderr,
        )
        return CollaborationConfig.get_example_config()
    with path.open() as f:
        return CollaborationConfig.load(f)


def _verify_directory(raw: str) -> pathlib.Path:
    ret = pathlib.Path(raw)
    if ret.exists() and not ret.is_dir():
        raise argparse.ArgumentTypeError(f"{ret} is a file, not a directory")
    return ret


def main(args: t.Optional[t.Sequence[t.Text]] = None) -> None:
    ap = get_argparse()
    namespace = ap.parse_args(args)
    execute_command(namespace)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Abstraction around efficient scaled matching of SignalTypes

This usually involves build an efficient datastructure for lookups
(aka, an index).

At scale, the flow for matching looks something like:
1. Fetch information from API, store it locally
2. Process local copy of the information, separate it out by
   which system will consume it
3. Build indices for distribution (or deltas to those indices if
   you are really fancy)
4. In the service that does matching, swap out the index to the
   new one.
5. Hash content at upload sites, check against the service hosting
   the index.

"""

import typing as t


T = t.TypeVar("T")


class IndexMatch(t.Generic[T]):
    """
    Wrapper around a match to the index, which may or may not
    be an actual match based on the distance settings for a source.

    The index is free to optimize not returning values that are
    greater than would be considered matches by a downstream (i.e.
    31 is probably the max distance that would be used in PDQ)

    NamedTuple can't be made generic, so here's an equivalent
    """

    __slots__ = ["distance", "metadata"]
    distance: int
    metadata: T

    def __init__(self, distance: int, metadata: T) -> None:
        self.distance = distance
        self.metadata = metadata


class SignalTypeIndex(t.Generic[T]):
    """
    Abstraction for an efficient matching technique.

    This class can be thought of as just a Dict[hash, List[T]] interface
    that can return the flattened result of multiple entries on a get().

    The T can be whatever metadata might be useful, although be warned
    that not all T may serialize correctly.

    # Why is `hash` str?
    There are various multi-pass hashing approaches that may generate more
    complicated data structures, or require multiple query passes. It's
    unclear whether we should try and generalize those yet, and forcing the
    type to always be serializable in str gives us the advantage that
    uploading it to ThreatExchange is always straightforward.

    # Handling Restricted Value Types
    In cases where your underlying index has limited type support for
    values (i.e. FAISS seems to only support ints), the easiest workaround
    is to just generate your own internal ids and maintain a separate map of
    the real results.

    fancy_index_only_to_int[str, int]
    internal_mapping List[T] or Dict[int, T]

    def get(hash: str):
        matches = fancy_index_only_to_int[str]
        ret = []
        for idx in matches:
            entry = internal_mapping[idx]
            ret.append(entry)
        return ret

    def add(hash: str, meta: T):
        idx = fancy_index_only_to_int.put(hash)
        internal_mapping[idx] = meta
    """

    def query(self, hash: str) -> t.List[IndexMatch[T]]:
        """
        Look up entries against the index, up to the max supported distance.

        The index is free to optimize not returning values that are
        greater than would be considered matches by a downstream system
        (i.e. 31 is probably the max distance that would be used in PDQ)
        """
        raise NotImplementedError

    @classmethod
    def build(cls, entries: t.Iterable[t.Tuple[str, T]]) -> "SignalTypeIndex[T]":
        """
        Build an index from a set of entries.

        Note that there may be duplicates of the hash type, i.e.

        H1: M1
        H1: M2

        A later call to query(H1) should return both M1 and M2
        """
        raise NotImplementedError

    def add(cls, entries: t.Iterable[t.Tuple[str, T]]) -> None:
        """
        Add entries to an existing index. May contain elements already in the
        index.
        """
        raise NotImplementedError

    def serialize(self, fout: t.BinaryIO) -> None:
        """
        Convert the index into a bytestream (probably a file).

        The reason to use an IO instead of just stringifying it is that some
        indices might be really big, and we don't want to pay a cost of 2x-ing
        our memory storage (versus writing to file).

        Could also be premature optimization, you decide!
        """
        raise NotImplementedError

    @classmethod
    def deserialize(cls, fin: t.BinaryIO) -> "SignalTypeIndex[T]":
        """Instanciate an index from a previous call to serialize"""
        raise NotImplementedError



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Wrapper around the Photo PDQ signal type.
"""

import typing as t
import pathlib
import warnings

from . import signal_base
from ..hashing.pdq_utils import pdq_match, BITS_IN_PDQ


def _raise_pillow_warning():
    warnings.warn(
        "PDQ from raw image data requires Pillow and pdqhash to be installed; install threatexchange with the [pdq_hasher] extra to use them",
        category=UserWarning,
    )


class PdqSignal(
    signal_base.SimpleSignalType, signal_base.FileHasher, signal_base.BytesHasher
):
    """
    PDQ is an open source photo similarity algorithm.

    Unlike MD5s, which are sensitive to single pixel differences, PDQ has
    a concept of "distance" and can detect when content is visually similar.
    This property tends to make it much more effective at finding images that
    a human would claim are the same, but also opens the door for false
    positives.

    Which distance to use can differ based on the type of content being
    searched for. While the PDQ documentation suggests certain thresholds,
    they can sometimes vary depending on what you are comparing against.
    """

    INDICATOR_TYPE = "HASH_PDQ"
    TYPE_TAG = "media_type_photo"

    # This may need to be updated (TODO make more configurable)
    # Hashes of distance less than or equal to this threshold are considered a 'match'
    PDQ_CONFIDENT_MATCH_THRESHOLD = 31

    @classmethod
    def hash_from_file(cls, file: pathlib.Path) -> str:
        try:
            from threatexchange.hashing.pdq_hasher import pdq_from_file
        except:
            _raise_pillow_warning()
            return ""
        pdq_hash, _quality = pdq_from_file(file)
        return pdq_hash

    def match_hash(self, signal_str: str) -> t.List[signal_base.SignalMatch]:

        # for case where cli tries to match against non-pdq type hashes
        # (filtering should likely be moved up in future to avoid silent errors)
        if len(signal_str) != BITS_IN_PDQ / 4:
            return []

        return [
            signal_base.SignalMatch(signal_attr.labels, signal_attr.first_descriptor_id)
            for pdq_hash, signal_attr in self.state.items()
            if pdq_match(pdq_hash, signal_str, self.PDQ_CONFIDENT_MATCH_THRESHOLD)
        ]

    @classmethod
    def hash_from_bytes(self, bytes_: bytes) -> str:
        try:
            from threatexchange.hashing.pdq_hasher import pdq_from_bytes
        except:
            _raise_pillow_warning()
            return ""
        pdq_hash, quality = pdq_from_bytes(bytes_)
        return pdq_hash



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
import warnings

from threatexchange.hashing.pdq_utils import BITS_IN_PDQ

try:
    from threatexchange.hashing.pdq_faiss_matcher import (
        PDQHashIndex,
        PDQFlatHashIndex,
        PDQMultiHashIndex,
    )
except:
    warnings.warn(
        "pdq matchers require faiss to be installed; install threatexchange with the [faiss] extra to use them",
        category=ImportWarning,
    )



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Match command for parsing simple data sources against the dataset.
"""

import argparse
import pathlib
import sys
import typing as t

from ..api import ThreatExchangeAPI
from ..content_type import meta
from ..dataset import Dataset
from ..descriptor import ThreatDescriptor
from ..signal_type.signal_base import FileMatcher, HashMatcher, StrMatcher
from . import command_base, fetch


class MatchCommand(command_base.Command):
    """
    Match content to items in ThreatExchange.

    Using the dataset from the fetch command, try to match content. Not all
    content and hashing types are implemented, so it's possible that you
    can download signals, but not match them via this command. In some cases
    the implementation in this package is sub-optimal, either in completeness
    (i.e. only matching exact when near-matching is supported), or in runtime
    (i.e. using a linear implementation when a sublinear implementation exists)

    The output of this command is in the following format:

      <matched descriptor id> <signal type> <label1> <label2...>

    If tying this into your own integrity systems, if the result of this match
    is human review, you'll want to store the matched descriptor id and make
    a call to

      all_in_one label descriptor <matched descriptor id>

    with the results of that review.
    """

    USE_STDIN = "-"

    @classmethod
    def init_argparse(cls, ap) -> None:

        ap.add_argument(
            "content_type",
            choices=[t.get_name() for t in meta.get_all_content_types()],
            help="what kind of content to match",
        )

        ap.add_argument(
            "--hashes",
            "-H",
            action="store_true",
            help=(
                "instead of content (i.e. videos), "
                "input contains intermediate representations (i.e. video MD5s)"
            ),
        )

        ap.add_argument(
            "--as-text",
            "-T",
            action="store_true",
            help="force input to be interpreted as text instead of as filenames",
        )

        ap.add_argument(
            "content",
            nargs="+",
            help=(
                "what to match against. Accepts filenames, "
                "quoted strings, or '-' to read newline-separated stdin"
            ),
        )

        ap.add_argument(
            "--show-false-positives",
            action="store_true",
            help="show matches even if you've marked them false_positive",
        )

        ap.add_argument(
            "--hide-disputed",
            action="store_true",
            help="hide matches if someone has disputed them",
        )

    def __init__(
        self,
        content_type: str,
        hashes: bool,
        as_text: bool,
        content: t.List[str],
        show_false_positives: bool,
        hide_disputed: bool,
    ) -> None:
        self.content_type = [
            c for c in meta.get_all_content_types() if c.get_name() == content_type
        ][0]
        self.input_generator = self.parse_input(content, hashes, as_text)
        self.as_hashes = hashes
        self.show_false_positives = show_false_positives
        self.hide_disputed = hide_disputed

    def parse_input(
        self,
        input_: t.Iterable[str],
        input_is_hashes: bool,
        force_input_to_text: bool,
        no_stderr=False,
    ) -> t.Generator[t.Union[str, pathlib.Path], None, None]:
        def interpret_token(
            tok: str,
        ) -> t.Generator[t.Union[str, pathlib.Path], None, None]:
            if force_input_to_text:
                return tok
            path = pathlib.Path(token)
            if path.exists():
                return path
            return tok

        for token in input_:
            token = token.rstrip()
            if not no_stderr and token == self.USE_STDIN:
                yield from self.parse_input(
                    sys.stdin, input_is_hashes, force_input_to_text, no_stderr=True
                )
                continue
            parsed = interpret_token(token)
            if input_is_hashes and isinstance(parsed, pathlib.Path):
                yield from self.parse_input(
                    parsed.open("r"),
                    input_is_hashes=True,
                    force_input_to_text=True,
                    no_stderr=True,
                )
            else:
                yield parsed

    def execute(self, api: ThreatExchangeAPI, dataset: Dataset) -> None:
        if dataset.is_cache_empty:
            self.stderr(
                "Looks like you are running this for the first time. Fetching some sample data."
            )
            fetch.TagFetchCommand(sample=True).execute(api, dataset)

        all_signal_types = dataset.load_cache(
            s() for s in self.content_type.get_signal_types()
        )

        file_matchers = [s for s in all_signal_types if isinstance(s, FileMatcher)]
        str_matchers = [s for s in all_signal_types if isinstance(s, StrMatcher)]

        match_str = lambda s, t: s.match(t)
        if self.as_hashes:
            match_str = lambda s, t: s.match_hash(t)
            str_matchers = [s for s in all_signal_types if isinstance(s, HashMatcher)]

        seen = set()
        for inp in self.input_generator:
            match_fn = lambda s, t: s.match_file(t)
            signal_types = file_matchers
            if isinstance(inp, str):
                match_fn = match_str
                signal_types = str_matchers

            for signal_type in signal_types:
                for match in match_fn(signal_type, inp):
                    if match.primary_descriptor_id in seen:
                        continue
                    seen.add(match.primary_descriptor_id)
                    labels = sorted(
                        l
                        for l in match.labels
                        if l in dataset.config.labels_for_collaboration
                    )
                    # If a lone DISPUTED, this means it is just a lone NON_MALICIOUS
                    # No one does this intentionally, it means that it was originally
                    # a disputed descriptor where the original author withdrew later
                    # If they owner of the dangling NON_MALICIOUS descriptor had
                    # used a reaction, it would have cleaned itself up, so lets
                    # just not show it.
                    # (This can be overridded in the collab config e.g.
                    # for sample data we do want to show these descriptor.)
                    if (
                        labels == [ThreatDescriptor.DISPUTED]
                        and not dataset.config.show_safe_list
                    ):
                        continue
                    if self.hide_disputed and ThreatDescriptor.DISPUTED in labels:
                        continue
                    if (
                        not self.show_false_positives
                        and ThreatDescriptor.FALSE_POSITIVE in labels
                    ):
                        continue

                    print(
                        match.primary_descriptor_id,
                        signal_type.get_name(),
                        " ".join(labels),
                    )



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Implementation of SignalTypeIndex abstraction for PDQ by wrapping
hashing.pdq_faiss_matcher.
"""

from collections import OrderedDict
import collections
import typing as t
import pickle

from threatexchange.signal_type.index import SignalTypeIndex, IndexMatch, T as IndexT
from threatexchange.hashing.pdq_faiss_matcher import PDQMultiHashIndex


class PDQIndex(SignalTypeIndex):
    """
    Wrapper around the pdq faiss index lib
    """

    PDQ_CONFIDENT_MATCH_THRESHOLD = 31
    T = IndexT

    def __init__(self, entries: t.Iterable[t.Tuple[str, T]]) -> None:
        super().__init__()
        self.local_id_to_entry: t.OrderedDict = collections.OrderedDict()
        self.index = PDQMultiHashIndex()
        self.add(entries=entries)

    def __len__(self) -> int:
        return len(self.local_id_to_entry)

    def query(self, hash: str) -> t.List[IndexMatch[T]]:
        """
        Look up entries against the index, up to the max supported distance.
        """

        # query takes a signal hash but index supports banch quries hence [hash]
        results = self.index.search(
            [hash], self.PDQ_CONFIDENT_MATCH_THRESHOLD, return_as_ids=True
        )
        matches = []
        for result_ids in results:
            for id in result_ids:
                # distance = -1 (index does not currently support distance)
                matches.append(IndexMatch(-1, self.local_id_to_entry[id][1]))
        return matches

    def add(self, entries: t.Iterable[t.Tuple[str, T]]) -> None:
        hashes = []

        for i, entry in enumerate(entries):
            self.local_id_to_entry[i] = entry
            hashes.append(entry[0])

        self.index.add(hashes, self.local_id_to_entry.keys())

    @classmethod
    def build(cls, entries: t.Iterable[t.Tuple[str, T]]) -> "SignalTypeIndex[T]":
        """
        Build an PDQ index from a set of entries.
        """
        return cls(entries)

    def serialize(self, fout: t.BinaryIO) -> None:
        """
        Convert the PDQ index into a bytestream (probably a file).
        """
        fout.write(pickle.dumps(self))

    @classmethod
    def deserialize(cls, fin: t.BinaryIO) -> "SignalTypeIndex[T]":
        """
        Instanciate an index from a previous call to serialize
        """
        return pickle.loads(fin)



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Wrapper around the pdf content type.
"""
import typing as t

from ..signal_type import tlsh_pdf
from ..signal_type.signal_base import SignalType
from .content_base import ContentType


class PDFContent(ContentType):
    """
    Content that represents text in Portable Document Format.

    Examples might be:
    * PDFs
    """

    @classmethod
    def get_signal_types(cls) -> t.List[t.Type[SignalType]]:
        return [tlsh_pdf.TLSHSignal]



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Hash command to convert content into signatures.
"""

import argparse
import pathlib
import sys
import typing as t

from ..api import ThreatExchangeAPI
from ..content_type import meta
from ..dataset import Dataset
from ..descriptor import ThreatDescriptor
from ..signal_type.signal_base import FileHasher, StrHasher
from . import command_base, fetch


# TODO consider refactor to handle overlap with match
class HashCommand(command_base.Command):
    """
    Hash content into signatures (aka hashes).

    Reads inputs as filenames by default, or as text with --as-text.

    You can also pass in via stdin by using "-" as the input.
    """

    USE_STDIN = "-"

    @classmethod
    def init_argparse(cls, ap) -> None:

        ap.add_argument(
            "content_type",
            choices=[t.get_name() for t in meta.get_all_content_types()],
            help="what kind of content to hash",
        )

        ap.add_argument(
            "--signal-type",
            "-S",
            choices=[t.get_name() for t in meta.get_all_signal_types()],
            help="only generate these signal types",
        )

        ap.add_argument(
            "--as-text",
            "-T",
            action="store_true",
            help="force input to be interpreted as text instead of as filenames",
        )

        ap.add_argument(
            "content",
            nargs="+",
            help=(
                "what to match against. Accepts filenames, "
                "quoted strings, or '-' to read newline-separated stdin"
            ),
        )

    def __init__(
        self,
        content_type: str,
        signal_type: t.Optional[str],
        as_text: bool,
        content: t.List[str],
    ) -> None:
        self.content_type = [
            c for c in meta.get_all_content_types() if c.get_name() == content_type
        ][0]
        self.signal_type = signal_type

        if content == [self.USE_STDIN]:
            content = sys.stdin
        self.input_generator = self._parse_input(content, as_text)

    def _parse_input(
        self,
        input_: t.Iterable[str],
        force_input_to_text: bool,
    ) -> t.Generator[t.Union[str, pathlib.Path], None, None]:
        for token in input_:
            token = token.rstrip()
            if force_input_to_text:
                yield token
            else:
                yield pathlib.Path(token)

    def execute(self, api: ThreatExchangeAPI, dataset: Dataset) -> None:

        all_signal_types = [
            s
            for s in self.content_type.get_signal_types()
            if self.signal_type in (None, s.get_name())
        ]

        file_hashers = [s for s in all_signal_types if issubclass(s, FileHasher)]
        str_hashers = [s for s in all_signal_types if issubclass(s, StrHasher)]

        for inp in self.input_generator:
            hash_fn = lambda s, t: s.hash_from_file(t)
            signal_types = file_hashers
            if isinstance(inp, str):
                hash_fn = lambda s, t: s.hash_from_str(t)
                signal_types = str_hashers

            for signal_type in signal_types:
                hash_str = hash_fn(signal_type, inp)
                if hash_str:
                    print(signal_type.get_name(), hash_str)



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Abstraction for different content types.

This records all the valid signal types for a piece of content.
"""

import typing as t

from .. import common
from ..signal_type import signal_base, raw_text, trend_query, url


class ContentType:
    @classmethod
    def get_name(cls) -> str:
        """The human-friendly display name"""
        return common.class_name_to_human_name(cls.__name__, "Content")

    @classmethod
    def get_signal_types() -> t.List[t.Type[signal_base.SignalType]]:
        raise NotImplementedError



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

import argparse
import collections
import csv
import sys
import typing as t

from .. import threat_updates
from ..api import ThreatExchangeAPI
from ..content_type import meta
from ..dataset import Dataset
from . import command_base
from .dataset.simple_serialization import CliIndicatorSerialization


class DatasetCommand(command_base.Command):
    """
    Introspect the local ThreatExchange dataset.

    Can print out contents in simple formats
    (ideal for sending to another system), or regenerate
    index files (ideal if distributing indices for some reason)
    """

    @classmethod
    def init_argparse(cls, ap) -> None:
        ap.add_argument(
            "--rebuild-indices",
            "-r",
            action="store_true",
            help="rebuild indices",
        )
        ap.add_argument(
            "--print",
            "-p",
            action="store_true",
            dest="print_records",
            help="print records to screen",
        )
        ap.add_argument(
            "--type",
            "-t",
            dest="only_type",
            metavar="STR",
            help="only process one type of indicator (eg HASH_MD5), signal (eg photo_md5), or content (eg photo); specifying a signal or content type implies '--signal-summary'",
        )
        ap.add_argument(
            "--tag",
            "-g",
            dest="only_tag",
            metavar="STR",
            help="only process a single tag; implies '--signal-summary'",
        )
        ap.add_argument(
            "--indicator-only", "-i", action="store_true", help="only print indicators"
        )
        ap.add_argument(
            "--signal-summary",
            "-s",
            action="store_true",
            help="print summary in terms of signals",
        )

    def __init__(
        self,
        rebuild_indices: bool,
        only_type: t.Optional[str],
        only_tag: t.Optional[str],
        indicator_only: bool,
        signal_summary: bool,
        print_records: bool,
    ) -> None:
        self.rebuild_indices = rebuild_indices
        self.only_type = only_type
        self.only_tag = only_tag
        self.indicator_only = indicator_only
        self.signal_summary = signal_summary
        self.print_records = print_records

    def execute(self, api: ThreatExchangeAPI, dataset: Dataset) -> None:
        stores = [
            threat_updates.ThreatUpdateFileStore(
                dataset.state_dir,
                privacy_group,
                api.app_id,
                serialization=CliIndicatorSerialization,
            )
            for privacy_group in dataset.config.privacy_groups
        ]
        indicators = {}
        for store in stores:
            indicators.update(store.load_state())

        if self.only_type:
            signal_types = []
            s_type = meta.get_signal_types_by_name().get(self.only_type)
            if s_type:
                signal_types.append(s_type)
                self.signal_summary = True

            content_type = meta.get_content_types_by_name().get(self.only_type)
            if content_type:
                signal_types.extend(content_type.get_signal_types())
                self.signal_summary = True

            indicators = {
                k: v
                for k, v in indicators.items()
                if v.indicator_type == self.only_type
                or (
                    signal_types
                    and any(
                        sig_type.indicator_applies(v.indicator_type, v.rollup.labels)
                        for sig_type in signal_types
                    )
                )
            }

        if self.only_tag:
            self.signal_summary = True
            indicators = {
                k: v for k, v in indicators.items() if self.only_tag in v.rollup.labels
            }

        if self.rebuild_indices:
            generate_cli_indices(dataset, stores)
        if self.print_records:
            self._print_records(indicators)
        elif self.signal_summary:
            self.print_signal_summary(indicators)
        else:
            self.print_summary(indicators)

    def print_summary(self, indicators: t.Dict[str, CliIndicatorSerialization]):
        by_type = collections.Counter()
        for indicator in indicators.values():
            by_type[indicator.indicator_type] += 1
        for name, count in sorted(by_type.items(), key=lambda i: -i[1]):
            self.stderr(f"{name}: {count}")

    def print_signal_summary(self, indicators: t.Dict[str, CliIndicatorSerialization]):
        signal_types = meta.get_signal_types_by_name()
        by_signal = collections.Counter()
        for indicator in indicators.values():
            for name, signal_type in signal_types.items():
                if signal_type.indicator_applies(
                    indicator.indicator_type, indicator.rollup.labels
                ):
                    by_signal[name] += 1
        for name, count in sorted(by_signal.items(), key=lambda i: -i[1]):
            self.stderr(f"{name}: {count}")

    def _print_records(self, indicators: t.Dict[str, CliIndicatorSerialization]):
        csv_writer = csv.writer(sys.stdout)
        for indicator in indicators.values():
            if self.indicator_only:
                print(indicator.indicator)
            else:
                csv_writer.writerow(indicator.as_csv_row())


def generate_cli_indices(dataset: Dataset, indicator_stores):
    signal_types = meta.get_signal_types_by_name()
    indicators = {name: [] for name in signal_types}
    for store in indicator_stores:
        for indicator in store.load_state().values():
            for name, signal_type in signal_types.items():
                if signal_type.indicator_applies(
                    indicator.indicator_type, indicator.rollup.labels
                ):
                    indicators[name].append(indicator)

    for name, signal_type in signal_types.items():
        index_cls = signal_type.get_index_cls()
        index = None
        indicators_for_signal = indicators[name]
        if indicators_for_signal:
            index = index_cls.build(
                (indicator.indicator_type, indicator.rollup)
                for indicator in indicators_for_signal
            )
        dataset.store_index(signal_type, index)



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
A wrapper around loading and storing ThreatExchange data from files.

There are a few categories of state that this wraps:
  1. Checkpoints - state about previous fetches
  2. Collaboration Indicator Dumps - Raw output from threat_updates
  3. Index state - serializations of indexes for SignalType
"""

import json
import pathlib
import typing as t

from . import collab_config
from .content_type import meta
from .signal_type import signal_base
from .signal_type import index


class FetchCheckpoint(t.NamedTuple):
    last_full_fetch: float
    last_fetch: float

    def next(self, fetch_start_time: float, full_fetch: bool) -> "FetchCheckpoint":
        full_fetch = full_fetch or not self.last_full_fetch
        return FetchCheckpoint(
            fetch_start_time if full_fetch else self.last_full_fetch, fetch_start_time
        )

    def serialize(self) -> str:
        return f"{self.last_full_fetch} {self.last_fetch}"

    @classmethod
    def deserialize(cls, s: str) -> "FetchCheckpoint":
        last_full, _, last = s.partition(" ")
        return cls(float(last_full), float(last))


class Dataset:

    EXTENSION = ".te"

    def __init__(
        self,
        config: collab_config.CollaborationConfig,
        state_dir: t.Optional[pathlib.Path] = None,
    ) -> None:
        self.config = config
        if state_dir is None:
            state_dir = pathlib.Path.home() / config.default_state_dir_name
            assert not state_dir.is_file()
        self.state_dir = state_dir

    @property
    def is_cache_empty(self) -> bool:
        return not (
            self.state_dir.exists() and any(self.state_dir.glob(f"*{self.EXTENSION}"))
        )

    def _fetch_checkpoint_path(self) -> pathlib.Path:
        return self.state_dir / f"fetch_checkpoint{self.EXTENSION}"

    def _indicator_checkpoint_path(self, privacy_group: int) -> pathlib.Path:
        return (
            self.state_dir / f"indicators/{privacy_group}/_checkpoint{self.EXTENSION}"
        )

    def clear_cache(self) -> None:
        for p in self.state_dir.iterdir():
            if p.suffix == self.EXTENSION:
                p.unlink()

    def record_fetch_checkpoint(
        self, fetch_started_timestamp: float, full_fetch: bool
    ) -> None:
        prev = self.get_fetch_checkpoint()
        with self._fetch_checkpoint_path().open("w+") as f:
            f.write(prev.next(fetch_started_timestamp, full_fetch).serialize())

    def get_fetch_checkpoint(self) -> FetchCheckpoint:
        checkpoint = self._fetch_checkpoint_path()
        if not checkpoint.exists():
            return FetchCheckpoint(0, 0)
        return FetchCheckpoint.deserialize(checkpoint.read_text())

    def _signal_state_file(self, signal_type: signal_base.SignalType) -> pathlib.Path:
        return self.state_dir / f"{signal_type.get_name()}{self.EXTENSION}"

    def _index_file(self, signal_type: signal_base.SignalType) -> pathlib.Path:
        return self.state_dir / f"{signal_type.get_name()}.index{self.EXTENSION}"

    def store_cache(self, signal_type: signal_base.SignalType) -> None:
        if not self.state_dir.exists():
            self.state_dir.mkdir()
        signal_type.store(self._signal_state_file(signal_type))

    def load_cache(
        self, signal_types: t.Optional[t.Iterable[signal_base.SignalType]] = None
    ) -> t.List[signal_base.SignalType]:
        """Load everything in the state directory and initialize signal types"""
        if signal_types is None:
            signal_types = [s() for s in meta.get_all_signal_types()]
        ret = []
        for signal_type in signal_types:
            signal_state_file = self._signal_state_file(signal_type)
            if signal_state_file.exists():
                signal_type.load(signal_state_file)
            ret.append(signal_type)
        return ret

    def store_index(self, signal_type: signal_base.SignalType, index) -> None:
        if not self.state_dir.exists():
            self.state_dir.mkdir()
        path = self._index_file(signal_type)
        if index is None:
            if path.exists():
                path.unlink()
            return
        with path.open("wb") as fout:
            index.serialize(fout)

    def load_index(
        self, signal_type: signal_base.SignalType
    ) -> t.Optional[index.SignalTypeIndex]:
        path = self._index_file(signal_type)
        if not path.exists():
            return None
        with path.open("rb") as fin:
            return signal_type.get_index_cls.deserialize(fin)



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Wrapper around the Trend Query (keywords and regexes) content type.
"""

import csv
import json
import pathlib
import re
import typing as t

from ..descriptor import SimpleDescriptorRollup, ThreatDescriptor
from . import signal_base


class TrendQuery:
    """
    A parsed trend query, based on regexes.
    """

    REGEX_PREFIX = "regex-"

    # re.pattern is not in 3.6, which is what we are targeting right now
    def __init__(self, query_json: t.Dict[str, t.Any]) -> None:
        self.and_terms: t.List[t.List[t.Any]] = [
            [self._parse_term(t) for t in and_["or"]] for and_ in query_json["and"]
        ]
        self.not_terms: t.List[t.Any] = [self._parse_term(t) for t in query_json["not"]]

    def _parse_term(self, t) -> t.Any:
        if t.startswith(self.REGEX_PREFIX):
            return re.compile(t[len(self.REGEX_PREFIX) + 1 : -1])
        return re.compile(f"\\b{re.escape(t)}\\b")

    def _match_term(self, t: t.Union[str, t.Any], text: str) -> bool:
        return bool()

    def matches(self, text: str) -> bool:
        for or_ in self.and_terms:
            if not any(t.search(text) for t in or_):
                break
        else:
            return not any(t.search(text) for t in self.not_terms)
        return False


class TrendQuerySignal(signal_base.SignalType, signal_base.StrMatcher):
    """
    Trend Queries are a combination of and/or/not regexes.

    Based off a system effective for grouping content at Facebook, Trend Queries
    can potentially help you sift though large sets of content to quickly flag
    ones that might be interesting to you.

    They have high "recall" but potentially low "precision".
    """

    def __init__(self) -> None:
        self.state: t.Dict[str, (TrendQuery, SimpleDescriptorRollup)] = {}

    def process_descriptor(self, descriptor: ThreatDescriptor) -> bool:
        """
        Add ThreatDescriptor to the state of this type, if it is for this type

        Return true if the true if the descriptor was used.
        """
        if (
            descriptor.indicator_type != "DEBUG_STRING"
            or "media_type_trend_query" not in descriptor.tags
        ):
            return False
        old_val = self.state.get(descriptor.raw_indicator)
        query_json = json.loads(descriptor.raw_indicator)

        if old_val is None:
            self.state[descriptor.raw_indicator] = (
                TrendQuery(query_json),
                SimpleDescriptorRollup(
                    descriptor.id, descriptor.added_on, descriptor.tags
                ),
            )
        else:
            old_val[1].merge(descriptor)
        return True

    def match(self, content: str) -> t.List[signal_base.SignalMatch]:
        return [
            signal_base.SignalMatch(rollup.labels, rollup.first_descriptor_id)
            for query, rollup in self.state.values()
            if query.matches(content)
        ]

    def load(self, path: pathlib.Path) -> None:
        self.state.clear()
        csv.field_size_limit(path.stat().st_size)  # dodge field size problems
        with path.open("r", newline="") as f:
            for row in csv.reader(f, dialect="excel-tab"):
                raw_indicator = row[0]
                query_json = json.loads(raw_indicator)
                self.state[raw_indicator] = (
                    TrendQuery(query_json),
                    SimpleDescriptorRollup.from_row(row[1:]),
                )

    def store(self, path: pathlib.Path) -> None:
        with path.open("w+", newline="") as f:
            writer = csv.writer(f, dialect="excel-tab")
            for k, v in self.state.items():
                writer.writerow((k,) + v[1].as_row())

    @classmethod
    def indicator_applies(cls, indicator_type: str, tags: t.List[str]) -> bool:
        return indicator_type == "DEBUG_STRING" and "media_type_trend_query" in tags



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Stores all the content type implementations for lookup
"""

import functools
import typing as t

from . import content_base, text, video, photo, pdf
from ..signal_type import signal_base


@functools.lru_cache(1)
def get_all_content_types() -> t.List[t.Type[content_base.ContentType]]:
    """Returns all content_type implementations for commands"""
    return [text.TextContent, video.VideoContent, photo.PhotoContent, pdf.PDFContent]


@functools.lru_cache(1)
def get_content_types_by_name() -> t.Dict[str, content_base.ContentType]:
    return {c.get_name(): c for c in get_all_content_types()}


@functools.lru_cache(1)
def get_all_signal_types() -> t.Set[t.Type[signal_base.SignalType]]:
    """Returns all signal_type implementations for commands"""
    ret = set()
    return {s for c in get_all_content_types() for s in c.get_signal_types()}


@functools.lru_cache(1)
def get_signal_types_by_name() -> t.Dict[str, signal_base.SignalType]:
    return {s.get_name(): s for s in get_all_signal_types()}


@functools.lru_cache(maxsize=None)
def _get_content_type_map():
    return {
        content_type.get_name(): content_type
        for content_type in get_all_content_types()
    }


def get_content_type_for_name(name: str) -> t.Type[content_base.ContentType]:
    """
    Given a name, get the ContentTYpe which returns that name on calling
    get_name().

    Note: Raises KeyError if not found.
    """
    return _get_content_type_map()[name]



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Wrapper around the URL signal type.
"""

import typing as t

from ..descriptor import SimpleDescriptorRollup, ThreatDescriptor
from . import signal_base


class URLSignal(signal_base.SimpleSignalType, signal_base.StrMatcher):
    """
    Wrapper around URL links, such as https://github.com/
    """

    # TODO - Also handle URI indicator_type
    INDICATOR_TYPE = "RAW_URI"
    TYPE_TAG = "media_type_url"

    def match(self, content) -> t.List[signal_base.SignalMatch]:
        ret = []
        for word in content.split():
            found = self.state.get(word)
            if found:
                ret.append(
                    signal_base.SignalMatch(found.labels, found.first_descriptor_id)
                )
        return ret



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Core abstractions for signal types.
"""

import csv
import pathlib
import pickle
import typing as t

from .. import common
from ..descriptor import SimpleDescriptorRollup, ThreatDescriptor
from . import index


class TrivialSignalTypeIndex(index.SignalTypeIndex):
    """
    Index that does only exact matches and serializes with pickle
    """

    def __init__(self) -> None:
        self.state: t.Dict[str, t.List[t.Any]] = {}

    def query(self, hash: str) -> t.List[index.IndexMatch[index.T]]:
        return [index.IndexMatch(0, meta) for meta in self.state.get(hash, [])]

    def add(self, vals: t.Iterable[t.Tuple[str, t.Any]]) -> None:
        for k, val in vals:
            l = self.state.get(k)
            if not l:
                l = []
                self.state[k] = l
            l.append(val)

    @classmethod
    def build(cls, vals: t.Iterable[t.Tuple[str, t.Any]]):
        ret = cls()
        ret.add(vals=vals)
        return ret

    def serialize(self, fout: t.BinaryIO):
        pickle.dump(self, fout)

    @classmethod
    def deserialize(cls, fin: t.BinaryIO):
        return pickle.load(fin)


class SignalMatch(t.NamedTuple):
    # TODO - Labels probably don't belong here, because we just duplicate storage
    #        better to have descriptor lookup with the full labels
    labels: t.Set[str]
    primary_descriptor_id: int
    # TODO: Someday, also include the following:
    # contested: bool
    # plurality_opinion: t.Tuple(bool, t.Set[str])
    # highest_action_severity: str
    # opinions_by_owner: t.Dict[int, MatchOpinion]


class SignalType:
    """
    Abstraction for different signal types.

    A signal type is an intermediate representation of content that can be used
    to match against similar or identical content. Sometimes called a "hash"
    type.

    This class additionally helps translates ThreatDescriptors into the correct
    representation to do matching, as well as serialize that representation
    into a compact form.
    """

    @classmethod
    def get_name(cls):
        """A compact name in lower_with_underscore style (used in filenames)"""
        return common.class_name_to_human_name(cls.__name__, "Signal")

    @classmethod
    def get_index_cls(cls) -> t.Type[index.SignalTypeIndex]:
        """Return the index class that handles this signal type"""
        return TrivialSignalTypeIndex

    @classmethod
    def indicator_applies(cls, indicator_type: str, tags: t.List[str]) -> bool:
        """Does this indicator correspond to this signal type?"""
        raise NotImplementedError

    @classmethod
    def compare_hash(cls, hash1: str, hash2: str) -> int:
        """
        Compare the distance of two hashes, the key operaiton for matching.

        If the algorithm doesn't support distance, having 0 = match,
        1 = no hash.

        Note that this can just be a reference/helper, and the efficient
        version of the algorithm can live in the index class.
        """
        raise NotImplementedError

    ##########################################################################
    # TODO - Remove  these methods after refactor
    def process_descriptor(self, descriptor: t.Dict[str, t.Any]) -> bool:
        """
        Add ThreatDescriptor to the state of this type, if it is for this type.

        Return true if the true if the descriptor was used.
        """
        return False

    def load(self, path: pathlib.Path) -> None:
        raise NotImplementedError

    def store(self, path: pathlib.Path) -> None:
        raise NotImplementedError


class HashMatcher:
    def match_hash(self, hash: str) -> t.List[SignalMatch]:
        """
        Given a representation of this SignalType, return matches.

        Example - PDQ distance comparison, or MD5 exact comparison
        """
        raise NotImplementedError


class FileMatcher:
    def match_file(self, file: pathlib.Path) -> t.List[SignalMatch]:
        """
        Given a file containing content, return matches.
        """
        raise NotImplementedError


class BytesMatcher:
    def match_bytes(self, bytes_: bytes) -> t.List[SignalMatch]:
        """
        If you have already brought the file to memory, don't write it to disk,
        hash it right there.
        """
        raise NotImplementedError


class StrMatcher(FileMatcher):
    def match(self, content: str) -> t.List[SignalMatch]:
        """
        Given a string representation of content, return matches.

        You don't need to implement this if it doesn't make sense for your
        signal type.
        """
        raise NotImplementedError

    def match_file(self, path: pathlib.Path) -> t.List[SignalMatch]:
        return self.match(path.read_text())


class StrHasher(HashMatcher, StrMatcher):
    """
    This class can turn text into intermediary representations (hashes)
    """

    @classmethod
    def hash_from_str(cls, content: str) -> str:
        """Get a string representation of the hash from a string"""
        raise NotImplementedError

    def match(self, content: str) -> t.List[SignalMatch]:
        str_hash = self.hash_from_str(content)
        return self.match_hash(str_hash)


class FileHasher(HashMatcher, FileMatcher):
    """
    This class can hash files.

    If also inheiriting from StrHasher, put this second in the inheiretence
    to prefer file hashing to reading the file in as a Str.
    """

    @classmethod
    def hash_from_file(self, file: pathlib.Path) -> str:
        """Get a string representation of the hash from a file"""
        raise NotImplementedError

    def match_file(self, path: pathlib.Path) -> t.List[SignalMatch]:
        file_hash = self.hash_from_file(path)
        return self.match_hash(file_hash)


class BytesHasher(HashMatcher, BytesMatcher):
    """
    This class can hash bytes.
    """

    @classmethod
    def hash_from_bytes(self, bytes_: bytes) -> str:
        """Get a string representation of the hash from bytes."""
        raise NotImplementedError

    def match_bytes(self, bytes_: bytes) -> t.List[SignalMatch]:
        bytes_hash = self.hash_from_bytes(bytes_)
        return self.match_hash(bytes_hash)


class SimpleSignalType(SignalType, HashMatcher):
    """
    Dead simple implementation for loading/storing a SignalType.

    Assumes that the signal type can easily merge on a string.
    """

    INDICATOR_TYPE: t.Union[str, t.Tuple[str]] = ()
    TYPE_TAG: t.Optional[str] = None

    @classmethod
    def indicator_applies(cls, indicator_type: str, tags: t.List[str]) -> bool:
        types = cls.INDICATOR_TYPE
        if isinstance(cls.INDICATOR_TYPE, str):
            types = (cls.INDICATOR_TYPE,)
        if indicator_type not in types:
            return False
        if cls.TYPE_TAG is not None:
            return cls.TYPE_TAG in tags
        return True

    @classmethod
    def compare_hash(cls, hash1: str, hash2: str) -> int:
        if hash1 == hash2:
            return 0
        return 1

    ##########################################################################
    # TODO - Remove these methods after refactor
    def __init__(self) -> None:
        self.state: t.Dict[str, SimpleDescriptorRollup] = {}

    def process_descriptor(self, descriptor: ThreatDescriptor) -> bool:
        """
        Add ThreatDescriptor to the state of this type, if it is for this type

        Return true if the true if the descriptor was used.
        """
        if not self.indicator_applies(descriptor.indicator_type, descriptor.tags):
            return False
        old_val = self.state.get(descriptor.raw_indicator)
        if old_val is None:
            self.state[
                descriptor.raw_indicator
            ] = SimpleDescriptorRollup.from_descriptor(descriptor)
        else:
            old_val.merge(descriptor)
        return True

    def match_hash(self, signal_str: str) -> t.List[SignalMatch]:
        found = self.state.get(signal_str)
        if found:
            return [SignalMatch(found.labels, found.first_descriptor_id)]
        return []

    def load(self, path: pathlib.Path) -> None:
        self.state.clear()
        csv.field_size_limit(path.stat().st_size)  # dodge field size problems
        with path.open("r", newline="") as f:
            for row in csv.reader(f):
                self.state[row[0]] = SimpleDescriptorRollup.from_row(row[1:])

    def store(self, path: pathlib.Path) -> None:
        with path.open("w+", newline="") as f:
            writer = csv.writer(f)
            for k, v in self.state.items():
                writer.writerow((k,) + v.as_row())



##!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Common helpers and libraries for the all-in-one command.

Strongly consider moving classes and functions out of this file if it starts
to fill up.
"""

import argparse
import sys
import typing as t

from .. import common
from ..api import ThreatExchangeAPI
from ..dataset import Dataset


class CommandError(Exception):
    """Wrapper for exceptions which cause return codes"""

    def __init__(self, message: str, returncode: int = 1) -> None:
        super().__init__(message)
        self.returncode = returncode


class Command:
    """
    Simple wrapper around setting up commands for an argparse-based cli.
    """

    @classmethod
    def add_command_to_subparser(cls, subparsers) -> None:
        """
        Shortcut for adding the command to the parser.

        Propbably don't override.
        """
        command_ap = subparsers.add_parser(
            cls.get_name(),
            description=cls.get_description(),
            help=cls.get_help(),
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )
        command_ap.set_defaults(command_cls=cls)
        cls.init_argparse(command_ap)

    @classmethod
    def init_argparse(cls, argparse: argparse.ArgumentParser) -> None:
        """
        Program the command subparser for __init__

        Your argument names should match the argument names in __init__.
        Be careful of collisions with the top level arg names from all_in_one.py
        """
        pass

    @classmethod
    def get_name(cls) -> str:
        """The display name of the command"""
        return common.class_name_to_human_name(cls.__name__, "Command").replace(
            "_", "-"
        )

    @classmethod
    def get_description(self) -> str:
        """The long help of the command"""
        return self.__doc__

    @classmethod
    def get_help(cls) -> str:
        """The short help of the command"""
        line = cls.get_description().strip().partition("\n")[0]
        # Good luck debugging this! Slightly reformat short description
        # (toplevel --help)
        if line[0].isupper():
            first_word, sp, rem = line.partition(" ")
            line = f"{first_word.lower()}{sp}{rem}"
        if line[-1] == ".":
            line = line[:-1]
        return line

    @staticmethod
    def stderr(*args, **kwargs) -> None:
        """Convenience accessor to stderr"""
        print(*args, file=sys.stderr, **kwargs)

    def execute(self, api: ThreatExchangeAPI, dataset: Dataset) -> None:
        raise NotImplementedError



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
A wrapper around the collaboration config file.

This file controls the behavior of the all-in-one command, and roughly
corresponds to a single-signal sharing usecase.
"""

import json
import re
import typing as t

from . import descriptor


class CollaborationConfig:

    KEY_NAME = "name"
    KEY_PRIVACY_GROUPS = "privacy_groups"
    KEY_LABELS = "labels"
    KEY_SAMPLE_TAG = "sample_tag"

    def __init__(
        self,
        name: str,
        labels: t.Dict[str, t.Any],
        privacy_groups: t.List[int],
        sample_tag: t.Optional[str] = None,
        show_safe_list=False,  # (safe_list == signals only listed as NON_MALICIOUS, e.g. sample data)
    ):
        self.name = name
        self.privacy_groups = privacy_groups
        self.labels = labels
        self.sample_tag = sample_tag
        self.show_safe_list = show_safe_list

    @property
    def default_state_dir_name(self) -> str:
        return re.sub("\W+", "_", self.name.lower())

    @property
    def labels_for_collaboration(self) -> t.Set[str]:
        ret = set(self.labels)
        ret.update(descriptor.ThreatDescriptor.SPECIAL_TAGS)
        return ret

    @classmethod
    def load(cls, file: t.IO):
        content = json.load(file)
        return cls(
            content.get(cls.KEY_NAME, "Sample Config"),
            content[cls.KEY_LABELS],
            content.get(cls.KEY_PRIVACY_GROUPS, []),
            content.get(cls.KEY_SAMPLE_TAG),
        )

    def store(self, filename: str):
        with open(filename, "w") as f:
            out = {
                self.KEY_NAME: self.name,
                self.KEY_LABELS: self.labels,
                self.KEY_PRIVACY_GROUPS: self.privacy_groups,
            }
            if self.sample_tag:
                out[self.KEY_SAMPLE_TAG] = self.sample_tag
            json.dump(out, f, indent=4)

    @classmethod
    def get_example_config(cls) -> "CollabConfig":
        """
        Get a config to access the public sample data under media_priority_samples
        """
        return cls(
            name="ThreatExchange Samples",
            labels={"media_priority_samples": {}},
            privacy_groups=[],
            sample_tag="media_priority_samples",
            show_safe_list=True,  # Show sample data (lone NON_MALICIOUS descriptors)
        )



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Wrapper around the pdf signal type.
"""

import pathlib
import typing as t
import warnings
from io import StringIO
import tlsh
from pdfminer.converter import TextConverter
from pdfminer.layout import LAParams
from pdfminer.pdfdocument import PDFDocument
from pdfminer.pdfinterp import PDFResourceManager, PDFPageInterpreter
from pdfminer.pdfpage import PDFPage
from pdfminer.pdfparser import PDFParser

from ..descriptor import SimpleDescriptorRollup, ThreatDescriptor
from . import signal_base

TLSH_CONFIDENT_MATCH_THRESHOLD = 30
EXPECT_TLSH_HASH_LENGTH = 72
""" 
This is a new signal type not found on ThreatExchange, in order to test the implementation we
override this set of hashes we want to match with here.
 
ToDo: @BarrettOlson: make this override cleaner and configurable by signal type in the future 
"""
TEMP_MATCH_IMPLEMNTATION_CHECK_DB = [
    [
        "T145B2859FE708266211A3026277C7AEE5FF76402C636AD5BA2C2CC11C23A1F2957773D5",
        [["test"], 1],
    ]
]


class TLSHSignal(
    signal_base.SimpleSignalType, signal_base.FileHasher, signal_base.BytesHasher
):
    """
    Simple signal type for PDFs using TLSH

    Extracts the text from a given pdf using pdfminer.six and hashes it with TLSH

    """

    TYPE_TAG = "media_type_pdf"

    @classmethod
    def hash_from_file(cls, file: pathlib.Path) -> str:
        if not str(file).endswith(".pdf"):
            warnings.warn("File does not appear to be a pdf. ", category=UserWarning)
            return ""
        text = StringIO()
        with open(file, "rb") as in_file:
            parser = PDFParser(in_file)
            doc = PDFDocument(parser)
            rsrcmgr = PDFResourceManager()
            device = TextConverter(rsrcmgr, text, laparams=LAParams())
            interpreter = PDFPageInterpreter(rsrcmgr, device)
            for page in PDFPage.create_pages(doc):
                interpreter.process_page(page)
            return str(tlsh.hash(text.getvalue().encode()))

    def match_hash(self, signal_str: str) -> t.List[signal_base.SignalMatch]:
        matches = []
        if len(signal_str) == EXPECT_TLSH_HASH_LENGTH:
            for x in TEMP_MATCH_IMPLEMNTATION_CHECK_DB:
                if tlsh.diffxlen(x[0], signal_str) <= TLSH_CONFIDENT_MATCH_THRESHOLD:
                    matches.append(signal_base.SignalMatch(x[1][0], x[1][1]))
        return matches



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Wrapper around the video content type.
"""
import typing as t

from ..signal_type import md5, pdq, pdq_ocr
from ..signal_type.signal_base import SignalType
from .content_base import ContentType


class PhotoContent(ContentType):
    """
    Content that contains static visual imagery.

    Examples might be:
      * jpeg
      * png
      * gif (non-animated)
      * frames from videos
      * thumbnails of videos
    """

    @classmethod
    def get_signal_types(cls) -> t.List[t.Type[SignalType]]:
        return [md5.PhotoMD5Signal, pdq.PdqSignal, pdq_ocr.PdqOcrSignal]



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Wrapper around the raw text signal type.
"""

import math
import pathlib
import typing as t

import Levenshtein

from ..descriptor import SimpleDescriptorRollup, ThreatDescriptor
from .. import common
from . import signal_base


class RawTextSignal(signal_base.SimpleSignalType, signal_base.StrMatcher):
    """
    Raw text signal is the same as raw text content: the exact text content.

    Unlike other formats like photos or videos, it is difficult to come
    up with non-reversable hashes of text information which are also effective
    at detecting similar content.
    """

    INDICATOR_TYPE = "DEBUG_STRING"
    TYPE_TAG = "media_type_text"

    def __init__(self) -> None:
        super().__init__()
        self.normal_to_raw = {}

    def match(self, content: str) -> t.List[signal_base.SignalMatch]:
        return self.match_hash(content)

    def match_hash(self, signal_str: str) -> t.List[signal_base.SignalMatch]:
        normalized_str = common.normalize_string(signal_str)
        # Match considered if 95% match
        match_threshold = math.floor(len(normalized_str) * 0.05)
        matches = []
        for normalized_candidate, raw in self.normal_to_raw.items():
            ldiff = abs(len(normalized_candidate) - len(normalized_str))
            # Filter out anything that can't possibly match due to len difference
            # Could optimize this if needed by storing in length buckets/sorted by length
            # (What about text content fully contained in target?)
            if ldiff > match_threshold:
                continue
            distance = Levenshtein.distance(normalized_candidate, normalized_str)
            # Linear search for fun and profit (but not efficiency)
            if distance <= match_threshold:
                found = self.state[raw]
                matches.append(
                    signal_base.SignalMatch(found.labels, found.first_descriptor_id)
                )
        return matches

    def process_descriptor(self, descriptor: ThreatDescriptor) -> bool:
        if not super().process_descriptor(descriptor):
            return False
        self._postprocess_indicator(descriptor.raw_indicator)
        return True

    def _postprocess_indicator(self, indicator: str) -> None:
        normalized = common.normalize_string(indicator)
        self.normal_to_raw[common.normalize_string(normalized)] = indicator

    def load(self, path: pathlib.Path) -> None:
        super().load(path)
        for indicator in self.state:
            self._postprocess_indicator(indicator)



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

import sys
from codecs import open
from os import path

from setuptools import find_packages, setup


here = path.abspath(path.dirname(__file__))
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

with open(path.join(here, "DESCRIPTION.rst"), encoding="utf-8") as f:
    description = f.read()

with open(path.join(here, "version.txt"), encoding="utf-8") as f:
    version = f.read().strip()

extras_require = {
    "faiss": ["faiss-cpu>=1.6.3", "numpy"],
    "pdq_hasher": [
        "numpy",
        "pdqhash>=0.2.2",
        "Pillow",
    ],
    "ocr": [
        "pytesseract",
    ],
    "pdf": [
        "py-tlsh",
        "pdfminer.six",
    ],
}

all_extras = set(sum(extras_require.values(), []))
extras_require["test"] = sorted({"pytest"} | all_extras)
extras_require["package"] = ["wheel"]
extras_require["lint"] = ["black"]
extras_require["all"] = sorted(set(sum(extras_require.values(), [])))

setup(
    name="threatexchange",
    version=version,
    description="Python Library for Facebook ThreatExchange",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Facebook",
    author_email="threatexchange@fb.com",
    license="BSD",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3.6",
    ],
    keywords="facebook threatexchange",
    url="https://www.github.com/facebook/ThreatExchange",
    packages=find_packages(exclude=["tests*"]),
    install_requires=[
        "python-Levenshtein",
        "requests>=2.26.0",
        "urllib3>=1.26.0",  # For allow_methods
        "dataclasses",
        "python-dateutil",
    ],
    extras_require=extras_require,
    entry_points={"console_scripts": ["threatexchange = threatexchange.cli.main:main"]},
)



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Wrapper around the Photo PDQ plus OCR signal type.
Requires the installations of Optical character recognition (OCR) tools
"""

import math
import typing as t
import pathlib
import warnings

import Levenshtein

from ..descriptor import SimpleDescriptorRollup, ThreatDescriptor
from . import signal_base
from ..hashing.pdq_utils import pdq_match
from .. import common


class PdqOcrSignal(signal_base.SimpleSignalType, signal_base.FileHasher):
    """
    PDQ is an open source photo similarity algorithm. See 'pdq.py'
    This signal type combines pdq hashes with a text string found using
    optical character recognition (OCR) on the image hashed. This adds an
    additional 'distance' using levenshtein string comparison.

    Accuracy of matches will depend on quality of OCR implementations and selected thresholds.
    (ocr_utils uses the open source pytesseract wrapper of tesseract)
    """

    INDICATOR_TYPE = "HASH_PDQ_OCR"
    TYPE_TAG = "media_type_photo"

    # This may need to be updated (TODO make more configurable)
    # Hashes of distance less than or equal to this threshold are considered a 'match'
    PDQ_PLUS_OCR_CONFIDENT_MATCH_THRESHOLD = 31
    # Match considered if 90% of the strings match
    LEVENSHTEIN_DISTANCE_PERCENT_THRESHOLD = 0.10

    @classmethod
    def hash_from_file(cls, file: pathlib.Path) -> str:
        try:
            from ..hashing.pdq_hasher import pdq_from_file
            from ..hashing.ocr_utils import text_from_image_file
        except:
            warnings.warn(
                "Getting both PDQ hash and text of an image file using OCR requires additional libraries already be installed; install threatexchange with the [pdq_hasher & ocr] extra and see ocr_utils.py",
                category=UserWarning,
            )
            return []

        pdq_hash, quality = pdq_from_file(file)
        ocr_text = text_from_image_file(file)

        return f"{pdq_hash},{ocr_text}"

    def match_hash(self, signal_str: str) -> t.List[signal_base.SignalMatch]:
        content_pdq_hash, _, content_ocr_text = signal_str.partition(",")
        if not content_ocr_text:
            return []
        matches = []
        for pdq_hash_plus_ocr, signal_attr in self.state.items():
            te_pdq_hash, te_ocr_text = pdq_hash_plus_ocr.split(",", maxsplit=1)
            # PDQ Hash Match
            if pdq_match(
                te_pdq_hash,
                content_pdq_hash,
                self.PDQ_PLUS_OCR_CONFIDENT_MATCH_THRESHOLD,
            ):
                # Check for text match
                normalized_content_str = common.normalize_string(content_ocr_text)
                normalized_te_str = common.normalize_string(te_ocr_text)
                if self._levenshtein_text_match(
                    normalized_content_str,
                    normalized_te_str,
                    self.LEVENSHTEIN_DISTANCE_PERCENT_THRESHOLD,
                ):
                    matches.append(
                        signal_base.SignalMatch(
                            signal_attr.labels, signal_attr.first_descriptor_id
                        )
                    )
        return matches

    def _levenshtein_text_match(self, str_a: str, str_b: str, threshold: float) -> bool:
        """
        Returns true if strings match within the percent threshold using Levenshtein distance
        """
        # TODO find way to reuse similar code in raw_text.py without sacrificing performance.
        match_threshold = math.floor(len(str_a) * threshold)
        ldiff = abs(len(str_a) - len(str_b))
        # Filter out anything that can't possibly match due to len difference
        if ldiff > match_threshold:
            return False
        distance = Levenshtein.distance(str_a, str_b)
        return distance <= match_threshold



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Wrapper around the text content type.
"""
import typing as t

from ..signal_type import raw_text, trend_query, url
from ..signal_type.signal_base import SignalType
from .content_base import ContentType


class TextContent(ContentType):
    """
    Content that represents static blobs of text.

    Examples might be:
    * Posts
    * Profile descriptions
    * OCR from photos, if the text itself is the dominating element
      (i.e. a screenshot of a block of text)
    """

    @classmethod
    def get_signal_types(cls) -> t.List[t.Type[SignalType]]:
        return [raw_text.RawTextSignal, trend_query.TrendQuerySignal, url.URLSignal]



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
A place to put simple helpers that don't seem like they go anywhere else.

If this file starts getting large, break it up.
"""

import re
import unicodedata


def class_name_to_human_name(name: str, suffix: str) -> str:
    """Helper to make human-friendly names using a class name as a template"""
    if name.endswith(suffix):
        name = name[: -len(suffix)]
    return camel_case_to_underscore(name)


def camel_case_to_underscore(name: str) -> str:
    """
    Convert name in camel-case notation into lowercase+underscore notation.

    For example, AbcXyz will be converted into abc_xyz.
    """
    s1 = re.sub(r"(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub(r"([a-z0-9])([A-Z])", r"\1_\2", s1).lower()


def normalize_string(s: str) -> str:
    """
    Strip parts of the raw string to try and make matching more effective.

    There are many redundant parts of input strings, or parts that don't
    meaningfully contribute to whether its a match or not. Try and strip
    as much of that as possible.
    """
    # Lowercase
    # CrAzY cAsE => crazy case
    s = s.lower()
    # Strip accent characters
    # ãóë => aoe
    s = "".join(
        c for c in unicodedata.normalize("NFKD", s) if not unicodedata.combining(c)
    )
    # Strip repeats of 2+
    # w0000000t => w00t
    s = re.sub("(.)(\1)+", "\1\1", s)
    # Strip non alphanumerics (including spaces)
    #
    s = re.sub("[\W_]", "", s)
    return s



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

import argparse
import collections
import concurrent.futures
import csv
import datetime
import json
import os
import pathlib
import time
import typing as t

from .. import threat_updates
from ..api import ThreatExchangeAPI
from ..content_type import meta
from ..dataset import Dataset
from ..descriptor import SimpleDescriptorRollup
from ..signal_type import signal_base
from . import command_base
from . import dataset_cmd
from .dataset.simple_serialization import CliIndicatorSerialization


class FetchCommand(command_base.Command):
    """
    Download content from ThreatExchange to disk.

    Using the CollaborationConfig, download signals that
    correspond to a single collaboration, and store them in the state
    directory.

    This endpoint uses /threat_updates to fetch content sequentially, and in
    theory can be interrupted without issues.
    """

    PROGRESS_PRINT_INTERVAL_SEC = 30

    @classmethod
    def init_argparse(cls, ap) -> None:
        ap.add_argument(
            "--full",
            action="store_true",
            help="force a refetch from the beginning of time (this is almost certainly not needed)",
        )
        ap.add_argument(
            "--skip-index-rebuild",
            action="store_true",
            help="don't rebuild indices after fetch",
        )
        ap.add_argument("--limit", type=int, help="stop after fetching this many items")
        ap.add_argument(
            "--stop-time",
            type=int,
            help="only fetch until this point",
        )

    def __init__(
        self,
        full: bool,
        stop_time: t.Optional[int],
        limit: t.Optional[int],
        skip_index_rebuild: bool,
    ) -> None:
        self.full = full
        self.stop_time = stop_time
        self.limit = limit
        self.skip_index_rebuild = skip_index_rebuild

        # Progress
        self.current_pgroup = 0
        self.last_update_time = 0
        # Print first update after 5 seconds
        self.last_update_printed = time.time() - self.PROGRESS_PRINT_INTERVAL_SEC + 5
        self.processed = 0
        self.counts = collections.Counter()

    def execute(self, api: ThreatExchangeAPI, dataset: Dataset) -> None:
        privacy_groups = dataset.config.privacy_groups
        stores = []
        for privacy_group in privacy_groups:
            indicator_store = threat_updates.ThreatUpdateFileStore(
                dataset.state_dir,
                privacy_group,
                api.app_id,
                serialization=CliIndicatorSerialization,
            )
            stores.append(indicator_store)
            if self.full:
                indicator_store.reset()
            else:
                indicator_store.load_checkpoint()
            if indicator_store.stale:
                indicator_store.reset()
            self.last_update_time = indicator_store.fetch_checkpoint
            if len(privacy_groups) > 1:
                self.current_pgroup = privacy_group

            self._print_progress()
            if indicator_store.fetch_checkpoint >= time.time():
                continue

            delta = indicator_store.next_delta
            if self.stop_time:
                delta.end = self.stop_time
            try:
                delta.incremental_sync_from_threatexchange(
                    api, limit=self.limit, progress_fn=self._progress
                )
            except:
                self.stderr("Exception occurred! Attempting to save...")
                # Force delta to show finished
                delta.end = delta.current
                raise
            finally:
                if delta:
                    indicator_store.apply_updates(delta)

        self.stderr(f"Processed {self.processed} updates:")

        if self.processed:
            for name, count in sorted(self.counts.items(), key=lambda i: -i[1]):
                self.stderr(f"{name}: {count:+}")

        # Rebuild CLI indices
        if not self.skip_index_rebuild:
            self.stderr("Rebuilding match indices...")
            dataset_cmd.generate_cli_indices(dataset, stores)

    def _progress(self, update: threat_updates.ThreatUpdateJSON) -> None:
        self.processed += 1
        self.counts[update.threat_type] += -1 if update.should_delete else 1
        self.last_update_time = update.time

        now = time.time()
        if now - self.last_update_printed >= self.PROGRESS_PRINT_INTERVAL_SEC:
            self.last_update_printed = now
            self._print_progress()

    def _print_progress(self):
        processed = ""
        if self.processed:
            processed = f"Downloaded {self.processed} updates. "

        on_privacy_group = ""
        if self.current_pgroup:
            on_privacy_group = f"on PrivacyGroup({self.current_pgroup}) "

        from_time = ""
        if not self.last_update_time:
            from_time = "ages long past"
        elif self.last_update_time >= time.time():
            from_time = "moments ago"
        else:
            delta = datetime.datetime.utcfromtimestamp(
                time.time()
            ) - datetime.datetime.utcfromtimestamp(self.last_update_time)
            parts = []
            for name, div in (
                ("year", datetime.timedelta(days=365)),
                ("day", datetime.timedelta(days=1)),
                ("hour", datetime.timedelta(hours=1)),
                ("minute", datetime.timedelta(minutes=1)),
                ("second", datetime.timedelta(seconds=1)),
            ):
                val, delta = divmod(delta, div)
                if val or parts:
                    parts.append((val, name))

            from_time = "now"
            if parts:
                str_parts = []
                for val, name in parts:
                    if str_parts:
                        str_parts.append(f"{val:02}{name[0]}")
                    else:
                        s = "s" if val > 1 else ""
                        str_parts.append(f"{val} {name}{s} ")
                from_time = f"{''.join(str_parts).strip()} ago"

        self.stderr(
            f"{processed}Currently {on_privacy_group}at {from_time}",
        )



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Label command for uploading opinions (ThreatDescriptors) or reactions.
"""

import typing as t

from .. import descriptor
from ..api import ThreatExchangeAPI
from ..collab_config import CollaborationConfig
from ..content_type import meta, text
from ..dataset import Dataset
from . import command_base


class LabelCommand(command_base.Command):
    """
    Apply labels to items in ThreatExchange.

    Labeling descriptors as false_positive will cause them to stop triggering
    matches by default in the match command.

    Examples:
      # Label descriptor
      $> threatexchange -c te.cfg label false_positive,other_label descriptor 12345
    """

    @classmethod
    def init_argparse(cls, ap) -> None:
        ap.add_argument(
            "labels",
            type=lambda s: s.strip().split(","),
            metavar="CSV",
            help="labels to apply to item",
        )

        # TODO - Put the match command content logic in a common place, re-use it
        ap.add_argument(
            "content_type", choices=["descriptor"], help="what kind of content to label"
        )
        ap.add_argument("content", help="the content to label")

    def __init__(self, content_type: str, content: str, labels: t.List[str]) -> None:
        if content_type == "descriptor":
            self.descriptor_id = content

        # Remove any of the special tags that someone included for whatever reason
        self.labels = [
            l for l in labels if l not in descriptor.ThreatDescriptor.SPECIAL_TAGS
        ]
        self.false_positive_reaction = False
        # Only use reaction logic if not also adding true positive labels
        if descriptor.ThreatDescriptor.FALSE_POSITIVE in labels and not self.labels:
            self.false_positive_reaction = True

    def execute(self, api: ThreatExchangeAPI, dataset: Dataset) -> None:
        if not self.false_positive_reaction:
            raise NotImplementedError
        err_message, ex, response = api.react_to_threat_descriptor(
            self.descriptor_id, "DISAGREE_WITH_TAGS"
        )
        if ex:
            raise ex
        if err_message:
            raise command_base.CommandError(err_message)
        if not response or response.get("success", "true") == "false":
            raise command_base.CommandError(
                f"Mystery error - response says: {response}"
            )



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Wrapper around the MD5 signal types.
"""

import hashlib
import pathlib
import typing as t

from ..descriptor import SimpleDescriptorRollup, ThreatDescriptor
from . import signal_base


class VideoMD5Signal(
    signal_base.SimpleSignalType, signal_base.FileHasher, signal_base.BytesHasher
):
    """
    Simple signal type for Video MD5s.

    Videos are quite expensive to process due to their large size. A simple
    matching algorithm is to just match against the file MD5, since
    transcoding is expensive enough that many platforms don't bother doing it.

    Even a single pixel changes will generate a new MD5 - consider formats
    that are capable of some notion of similarity, such as TMK+PDQF.
    """

    INDICATOR_TYPE = "HASH_MD5"
    TYPE_TAG = "media_type_video"

    @classmethod
    def hash_from_file(cls, path: pathlib.Path) -> str:
        file_hash = hashlib.md5()
        blocksize = 8192
        with open(path, "rb") as f:
            chunk = f.read(blocksize)
            while chunk:
                file_hash.update(chunk)
                chunk = f.read(blocksize)
        return file_hash.hexdigest()

    @classmethod
    def hash_from_bytes(self, bytes_: bytes) -> str:
        bytes_hash = hashlib.md5()
        bytes_hash.update(bytes_)
        return bytes_hash.hexdigest()


class PhotoMD5Signal(VideoMD5Signal):
    """
    Simple signal type for Photo MD5s.

    Unlike Videos, transcoding of photos is quite common. This should be
    a format of last resort, as the open source PDQ algorithm will usually
    have much higher recall without too much loss in precision.
    """

    TYPE_TAG = "media_type_photo"
